# pylint: disable=no-init
# pylint: disable=old-style-class
# pylint: disable=too-few-public-methods
"""
Database ORM models managed by this Django app
Please do not integrate directly with these models!!!  This app currently
offers two APIs -- api.py for direct Python integration and receivers.py,
which leverages Django's signal framework.
"""

from __future__ import absolute_import, unicode_literals
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from model_utils.models import TimeStampedModel


@python_2_unicode_compatible
class Milestone(TimeStampedModel):
    """
    A Milestone is a representation of an accomplishment which can be
    attained by a user. Milestones have a base set of meta data
    describing the milestone, including id, name, and description.
    Milestones can be used to drive functionality and behavior 'behind
    the scenes' in Open edX, such as with the Pre-Requisite Course and
    Course Entrance Exam use cases.
    """
    namespace = models.CharField(max_length=255, db_index=True)
    name = models.CharField(max_length=255, db_index=True)
    display_name = models.CharField(max_length=255)
    description = models.TextField()
    active = models.BooleanField(default=True, db_index=True)

    class Meta:
        """ Meta class for this Django model """
        unique_together = (("namespace", "name"),)

    def __str__(self):
        return self.namespace


@python_2_unicode_compatible
class MilestoneRelationshipType(TimeStampedModel):
    """
    A MilestoneRelationshipType represents a category of link available
    between a Milestone and a particular learning object (such as a
    Course). In addition to learning objects, a MilestoneRelationshipType
    can also represent the link between a Milestone and other platform
    entities (such as a User).  For example, a Course Author may
    indicate that Course 101 "fulfills" Milestone A, creating a new
    CourseMilestone record in the process. When a User completes
    Course 101, a new UserMilestone record is created reflecting the
    newly-attained Milestone A.  The Course Author may also indicate
    that Course 102 "requires" Milestone A, yielding a second
    CourseMilestone record.  Because the User has gained Milestone A
    (via Course 101), they can access Course 102.

    This same process of indicating MilestoneRelationshipTypes can be
    applied to other learning objects as well, such as course content
    (XBlocks/modules).
    """
    # name = models.CharField(max_length=255, db_index=True, unique=True)

    name = models.CharField(max_length=25, db_index=True, unique=True)
    description = models.TextField(blank=True)
    active = models.BooleanField(default=True)

    def __str__(self):
        return self.name

    @classmethod
    # pylint: disable=invalid-name
    def get_supported_milestone_relationship_types(cls):
        """ The set of currently-allowed milestone relationship types (names) """
        RELATIONSHIP_TYPE_CHOICES = {
            'REQUIRES': 'requires',
            'FULFILLS': 'fulfills',
        }
        return RELATIONSHIP_TYPE_CHOICES


@python_2_unicode_compatible
class CourseMilestone(TimeStampedModel):
    """
    A CourseMilestone represents the link between a Course and a
    Milestone. Because Courses are not true Open edX entities (in the
    Django/ORM sense) the modeling and integrity will be limited to that
    of specifying CourseKeyFields in this model, as well as related ones
    below. In addition, a MilestoneRelationshipType specifies the
    particular sort of relationship that exists between the Course and
    the Milestone, such as "requires".
    """
    course_id = models.CharField(max_length=255, db_index=True)
    milestone = models.ForeignKey(Milestone, db_index=True)
    milestone_relationship_type = models.ForeignKey(MilestoneRelationshipType, db_index=True)
    active = models.BooleanField(default=True, db_index=True)

    class Meta:
        """ Meta class for this Django model """
        unique_together = (("course_id", "milestone"),)

    def __str__(self):
        return "%s:%s:%s" % (self.course_id, self.milestone_relationship_type, self.milestone)


@python_2_unicode_compatible
class CourseContentMilestone(TimeStampedModel):
    """
    A CourseContentMilestone represents the link between a specific
    Course module (such as an XBlock) and a Milestone. Because
    CourseContent objects are not true Open edX entities (in the
    Django/ORM sense) the modeling and integrity will be limited to that
    of specifying LocationKeyFields in this model, as well as related
    ones. In addition, a MilestoneRelationshipType specifies the
    particular sort of relationship that exists between the Milestone
    and the CourseContent, such as "requires" or "fulfills".
    """
    course_id = models.CharField(max_length=255, db_index=True)
    content_id = models.CharField(max_length=255, db_index=True)
    milestone = models.ForeignKey(Milestone, db_index=True)
    milestone_relationship_type = models.ForeignKey(MilestoneRelationshipType, db_index=True)
    requirements = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        help_text="Stores JSON data required to determine milestone fulfillment"
    )
    active = models.BooleanField(default=True, db_index=True)

    class Meta:
        """ Meta class for this Django model """
        unique_together = (("course_id", "content_id", "milestone"),)

    def __str__(self):
        return "%s:%s:%s" % (self.content_id, self.milestone_relationship_type, self.milestone)


@python_2_unicode_compatible
class UserMilestone(TimeStampedModel):
    """
    A UserMilestone represents an stage reached or event experienced
    by a User during their interactions with the Open edX platform.

    The use of the 'collected' field in this model could support future
    use cases such as "Goals", in which a User might keep a list of
    Milestones they are interested in attaining. Side Note: In the
    Mozilla Open Badges world, this collection concept is referred
    to as the user's "backpack".

    The 'source' field was originally introduced as a free-form auditing
    field to document the method, location, or event which triggered the
    collection of the milestone by this user.
    """
    user_id = models.IntegerField(db_index=True)
    milestone = models.ForeignKey(Milestone, db_index=True)
    source = models.TextField(blank=True)
    collected = models.DateTimeField(blank=True, null=True)
    active = models.BooleanField(default=True, db_index=True)

    class Meta:
        """ Meta class for this Django model """
        unique_together = ("user_id", "milestone")

    def __str__(self):
        return "%s:%s" % (self.user_id, self.milestone)



"""
A wrapper class around requested methods exposed in api.py
"""

from __future__ import absolute_import, unicode_literals
import types

from milestones import api as milestones_api


class MilestonesService(object):  # pylint: disable=too-few-public-methods
    """
    An xBlock service for xBlocks to talk to the Milestones subsystem.

    NOTE: This is a Singleton class. We should only have one instance of it!
    """

    _instance = None

    REQUESTED_FUNCTIONS = [
        'get_course_content_milestones',
    ]

    def __new__(cls, *args, **kwargs):
        """
        This is the class factory to make sure this is a Singleton
        """
        if not cls._instance:
            cls._instance = super(MilestonesService, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        """
        Class initializer, which just inspects the libraries and exposes the same functions
        listed in REQUESTED_FUNCTIONS
        """
        self._bind_to_requested_functions()

    def _bind_to_requested_functions(self):
        """
        bind module functions. Since we use underscores to mean private methods, let's exclude those.
        """
        for attr_name in self.REQUESTED_FUNCTIONS:
            attr = getattr(milestones_api, attr_name, None)
            if isinstance(attr, types.FunctionType) and not attr_name.startswith('_'):
                if not hasattr(self, attr_name):
                    setattr(self, attr_name, attr)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from __future__ import absolute_import
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('milestones', '0002_data__seed_relationship_types'),
    ]

    operations = [
        migrations.AddField(
            model_name='coursecontentmilestone',
            name='requirements',
            field=models.CharField(help_text=b'Stores JSON data required to determine milestone fulfillment', max_length=255, null=True, blank=True),
        ),
    ]



# pylint: disable=no-member,expression-not-assigned
"""
Application data management/abstraction layer.  Responsible for:

1) Accessing information from various resources:
* Internal application state  (through local models in models.py)
* External application state  (ORM bindings with other apps, yuck)
* Remote data services (through service adapters in resources.py)

2) Calculating derivative information from existing state:
* Algorithms and data manipulations
* Aggregations
* Annotations
* Alternative data representations

Accepts and returns standard Python data structures (dicts, arrays of dicts)
for easy consumption and manipulation by callers -- the queryset stops here!

Note the terminology difference at this layer vs. API -- create/fetch/update/delete/

When the time comes for remote resources, import the module like so:
if getattr(settings, 'TEST_MODE', False) or os.getenv('TRAVIS_MODE', False):
    import milestones.tests.mocks.resources as remote
else:
    import milestones.resources as remote
"""
from __future__ import absolute_import, unicode_literals

import six

from . import exceptions
from . import models as internal
from . import serializers


# PRIVATE/INTERNAL METHODS (public methods located further down)
def _get_milestone_relationship_type(relationship):
    """
    Retrieves milestone relationship type object from backend datastore
    """
    try:
        return internal.MilestoneRelationshipType.objects.get(
            name=relationship,
            active=True
        )
    except internal.MilestoneRelationshipType.DoesNotExist:
        exceptions.raise_exception(
            "MilestoneRelationshipType",
            {'name': relationship},
            exceptions.InvalidMilestoneRelationshipTypeException
        )


def _activate_record(record):
    """
    Enables database records by setting the 'active' attribute to True
    The queries in this module filter out inactive records as part of their criteria
    This effectively allows us to soft-delete records so they are not lost forever
    """
    record.active = True
    record.save()


def _inactivate_record(record):
    """
    Disables database records by setting the 'active' attribute to False
    The queries in this module filter out inactive records as part of their criteria
    This effectively allows us to soft-delete records so they are not lost forever
    """
    record.active = False
    record.save()


def _activate_milestone(milestone, propagate=True):
    """
    Activates an inactivated (soft-deleted) milestone as well as any inactive relationships
    """
    if propagate:
        [_activate_record(record) for record
         in internal.CourseMilestone.objects.filter(milestone_id=milestone.id, active=False)]

        [_activate_record(record) for record
         in internal.CourseContentMilestone.objects.filter(milestone_id=milestone.id, active=False)]

        [_activate_record(record) for record
         in internal.UserMilestone.objects.filter(milestone_id=milestone.id, active=False)]

    [_activate_record(record) for record
     in internal.Milestone.objects.filter(id=milestone.id, active=False)]


def _inactivate_milestone(milestone):
    """
    Inactivates an activated milestone as well as any active relationships
    """
    [_inactivate_record(record) for record
     in internal.CourseMilestone.objects.filter(milestone_id=milestone.id, active=True)]

    [_inactivate_record(record) for record
     in internal.CourseContentMilestone.objects.filter(milestone_id=milestone.id, active=True)]

    [_inactivate_record(record) for record
     in internal.UserMilestone.objects.filter(milestone_id=milestone.id, active=True)]

    [_inactivate_record(record) for record
     in internal.Milestone.objects.filter(id=milestone.id, active=True)]


# PUBLIC METHODS
def create_milestone(milestone, propagate=True):
    """
    Inserts a new milestone into app/local state given the following dictionary:
    {
        'name': string,
        'display_name': string,
        'namespace': string,
        'description': string
    }
    Returns an updated dictionary including a new 'id': integer field/value
    """
    # Trust, but verify...
    if not milestone.get('name'):
        exceptions.raise_exception("Milestone", milestone, exceptions.InvalidMilestoneException)
    if not milestone.get('namespace'):
        exceptions.raise_exception("Milestone", milestone, exceptions.InvalidMilestoneException)
    milestone_obj = serializers.deserialize_milestone(milestone)
    try:
        milestone = internal.Milestone.objects.get(
            namespace=milestone_obj.namespace,
            name=milestone_obj.name,
        )
        # If the milestone exists, but was inactivated, we can simply turn it back on
        if not milestone.active:
            _activate_milestone(milestone, propagate)
    except internal.Milestone.DoesNotExist:
        milestone = internal.Milestone.objects.create(
            namespace=milestone_obj.namespace,
            name=milestone_obj.name,
            description=milestone_obj.description,
            display_name=milestone_obj.display_name,
            active=True
        )
    return serializers.serialize_milestone(milestone)


def update_milestone(milestone):
    """
    Updates an existing milestone in app/local state
    Returns a dictionary representation of the object
    """
    milestone_obj = serializers.deserialize_milestone(milestone)
    try:
        milestone = internal.Milestone.objects.get(id=milestone_obj.id)
        milestone.name = milestone_obj.name
        milestone.namespace = milestone_obj.namespace
        milestone.description = milestone_obj.description
        milestone.active = milestone_obj.active
    except internal.Milestone.DoesNotExist:
        exceptions.raise_exception("Milestone", milestone, exceptions.InvalidMilestoneException)
    return serializers.serialize_milestone(milestone)


def delete_milestone(milestone):
    """
    Inactivates an existing milestone from app/local state
    No return currently defined for this operation
    """
    milestone_obj = serializers.deserialize_milestone(milestone)
    _inactivate_milestone(milestone_obj)


def fetch_milestone(milestone_id):
    """
    Retrieves a specific milestone from app/local state
    Returns a dictionary representation of the object
    """
    if not milestone_id:
        exceptions.raise_exception("Milestone", {'id': milestone_id}, exceptions.InvalidMilestoneException)
    milestone = {'id': milestone_id}
    milestones = fetch_milestones(milestone)
    if not milestones:
        exceptions.raise_exception("Milestone", milestone, exceptions.InvalidMilestoneException)
    return milestones[0]


def fetch_milestones(milestone):
    """
    Retrieves a set of matching milestones from app/local state
    Returns a list-of-dicts representation of the object
    """
    if milestone is None:
        exceptions.raise_exception("Milestone", milestone, exceptions.InvalidMilestoneException)
    milestone_obj = serializers.deserialize_milestone(milestone)
    if milestone_obj.id is not None:
        return serializers.serialize_milestones(internal.Milestone.objects.filter(
            id=milestone_obj.id,
            active=True,
        ))
    if milestone_obj.namespace:
        return serializers.serialize_milestones(internal.Milestone.objects.filter(
            namespace=six.text_type(milestone_obj.namespace),
            active=True
        ))

    # If we get to this point the caller is attempting to match on an unsupported field
    exceptions.raise_exception("Milestone", milestone, exceptions.InvalidMilestoneException)


def fetch_milestone_relationship_types():
    """
    Model accessor method to return supported milestone relationship types
    """
    return internal.MilestoneRelationshipType.get_supported_milestone_relationship_types()


def create_course_milestone(course_key, relationship, milestone):
    """
    Inserts a new course-milestone into app/local state
    No response currently defined for this operation
    """
    relationship_type = _get_milestone_relationship_type(relationship)
    milestone_obj = serializers.deserialize_milestone(milestone)
    try:
        relationship = internal.CourseMilestone.objects.get(
            course_id=six.text_type(course_key),
            milestone=milestone_obj,
            milestone_relationship_type=relationship_type
        )
        # If the relationship exists, but was inactivated, we can simply turn it back on
        if not relationship.active:
            _activate_record(relationship)
    except internal.CourseMilestone.DoesNotExist:
        relationship = internal.CourseMilestone.objects.create(
            course_id=six.text_type(course_key),
            milestone=milestone_obj,
            milestone_relationship_type=relationship_type,
            active=True
        )


def delete_course_milestone(course_key, milestone):
    """
    Removes an existing course-milestone from app/local state
    No response currently defined for this operation
    """
    try:
        relationship = internal.CourseMilestone.objects.get(
            course_id=six.text_type(course_key),
            milestone=milestone['id'],
            active=True,
        )
        _inactivate_record(relationship)
    except internal.CourseMilestone.DoesNotExist:
        # If we're being asked to delete a course-milestone link
        # that does not exist in the database then our work is done
        pass


def fetch_courses_milestones(course_keys, relationship=None, user=None):
    """
    Retrieves the set of milestones currently linked to the specified courses
    Optionally pass in 'relationship' (ex. 'fulfills') to filter down the set
    Optionally pass in 'user' to constrain the set to those which the user has collected
    """
    queryset = internal.CourseMilestone.objects.filter(
        course_id__in=course_keys,
        active=True
    ).select_related('milestone')

    # if milestones relationship type found then apply the filter
    if relationship is not None:
        mrt = _get_milestone_relationship_type(relationship)
        queryset = queryset.filter(
            milestone_relationship_type=mrt.id,
        )

    # To pull the list of milestones a user HAS, use get_user_milestones
    # Use fetch_courses_milestones to pull the list of milestones that a user does not yet
    # have for the specified course
    relationships = fetch_milestone_relationship_types()
    if relationship == relationships['REQUIRES'] and user and user.get('id', 0) > 0:
        queryset = queryset.exclude(
            milestone__usermilestone__in=internal.UserMilestone.objects.filter(user_id=user['id'], active=True)
        )

    return [serializers.serialize_milestone_with_course(milestone) for milestone in queryset]


def create_course_content_milestone(course_key, content_key, relationship, milestone, requirements=None):
    """
    Inserts a new course-content-milestone into app/local state
    No response currently defined for this operation
    """
    relationship_type = _get_milestone_relationship_type(relationship)
    milestone_obj = serializers.deserialize_milestone(milestone)
    requirements = serializers.serialize_requirements(requirements)
    try:
        relationship = internal.CourseContentMilestone.objects.get(
            course_id=six.text_type(course_key),
            content_id=six.text_type(content_key),
            milestone=milestone_obj,
            milestone_relationship_type=relationship_type
        )
        # If the relationship exists, but was inactivated, we can simply turn it back on
        if not relationship.active:
            relationship.requirements = requirements
            _activate_record(relationship)
        elif relationship.requirements != requirements:
            # Update requirements field if necessary
            relationship.requirements = requirements
            relationship.save()
    except internal.CourseContentMilestone.DoesNotExist:
        relationship = internal.CourseContentMilestone.objects.create(
            course_id=six.text_type(course_key),
            content_id=six.text_type(content_key),
            milestone=milestone_obj,
            milestone_relationship_type=relationship_type,
            requirements=requirements,
            active=True
        )


def delete_course_content_milestone(course_key, content_key, milestone):
    """
    Removes an existing course-content-milestone from app/local state
    No response currently defined for this operation
    """
    try:
        relationship = internal.CourseContentMilestone.objects.get(
            course_id=six.text_type(course_key),
            content_id=six.text_type(content_key),
            milestone=milestone['id'],
            active=True,
        )
        _inactivate_record(relationship)
    except internal.CourseContentMilestone.DoesNotExist:
        # If we're being asked to delete a course-content-milestone link
        # that does not exist in the database then our work is done
        pass


def fetch_course_content_milestones(content_key=None, course_key=None, relationship=None, user=None):
    """
    Retrieves the set of milestones currently linked to the specified course content
    Optionally pass in 'relationship' (ex. 'fulfills') to filter down the set
    Optionally pass in 'user' to further-filter the set (ex. for retrieving unfulfilled milestones)
    """
    queryset = internal.CourseContentMilestone.objects.filter(
        active=True
    ).select_related('milestone')

    if course_key is not None:
        queryset = queryset.filter(course_id=six.text_type(course_key))

    if content_key is not None:
        queryset = queryset.filter(content_id=six.text_type(content_key))

    if relationship is not None:
        mrt = _get_milestone_relationship_type(relationship)
        queryset = queryset.filter(milestone_relationship_type=mrt.id)

        # Filter for unfulfilled milestones for the given user
        if relationship == 'requires' and user and user.get('id'):
            queryset = queryset.exclude(
                milestone__usermilestone__in=internal.UserMilestone.objects.filter(user_id=user['id'], active=True)
            )

    return [serializers.serialize_milestone_with_course_content(ccm) for ccm in queryset]


def fetch_milestone_courses(milestone, relationship=None):
    """
    Retrieves the set of courses currently linked to the specified milestone
    Optionally pass in 'relationship' (ex. 'fulfills') to filter down the set
    """
    milestone_obj = serializers.deserialize_milestone(milestone)
    queryset = internal.CourseMilestone.objects.filter(
        milestone=milestone_obj,
        active=True
    ).select_related('milestone')

    # if milestones relationship type found then apply the filter
    if relationship is not None:
        queryset = queryset.filter(
            milestone_relationship_type__name=relationship,
        )

    return [serializers.serialize_milestone_with_course(milestone) for milestone in queryset]


def fetch_milestone_course_content(milestone, relationship=None):
    """
    Retrieves the set of course content modules currently linked to the specified milestone
    Optionally pass in 'relationship' (ex. 'fulfills') to filter down the set
    """
    milestone_obj = serializers.deserialize_milestone(milestone)
    queryset = internal.CourseContentMilestone.objects.filter(
        milestone=milestone_obj,
        active=True
    ).select_related('milestone')

    # if milestones relationship type found then apply the filter
    if relationship is not None:
        queryset = queryset.filter(
            milestone_relationship_type__name=relationship,
        )

    return [serializers.serialize_milestone_with_course_content(milestone) for milestone in queryset]


def create_user_milestone(user, milestone):
    """
    Inserts a new user-milestone into app/local state
    No response currently defined for this operation
    """
    milestone_obj = serializers.deserialize_milestone(milestone)
    try:
        relationship = internal.UserMilestone.objects.get(
            user_id=user['id'],
            milestone=milestone_obj.id,
        )
        # If the relationship exists, but was inactivated, we can simply turn it back on
        if not relationship.active:
            _activate_record(relationship)
    except internal.UserMilestone.DoesNotExist:
        relationship = internal.UserMilestone.objects.create(
            user_id=user['id'],
            milestone=milestone_obj,
            active=True
        )


def delete_user_milestone(user, milestone):
    """
    Removes an existing user-milestone from app/local state
    No response currently defined for this operation
    """
    try:
        record = internal.UserMilestone.objects.get(
            user_id=user['id'],
            milestone=milestone['id'],
            active=True,
        )
        _inactivate_record(record)
    except internal.UserMilestone.DoesNotExist:
        # If we're being asked to delete a user-milestone link
        # that does not exist in the database then our work is done
        pass


def fetch_user_milestones(user, milestone_data):
    """
    Retrieves the set of milestones currently linked to the specified user
    """
    queryset = internal.Milestone.objects.filter(
        usermilestone__user_id=user['id'],
        usermilestone__active=True,
    )

    # We don't currently support a 'fetch all' use case -- must supply at least one filter
    if not milestone_data.get('id') and not milestone_data.get('namespace'):
        exceptions.raise_exception("Milestone", milestone_data, exceptions.InvalidMilestoneException)

    if milestone_data.get('id'):
        queryset.filter(id=milestone_data['id'])

    if milestone_data.get('namespace'):
        queryset.filter(namespace=milestone_data['namespace'])

    return [serializers.serialize_milestone(milestone) for milestone in queryset]


def delete_content_references(content_key):
    """
    Inactivates references to content keys within this app (ref: api.py)
    Supports the 'delete entrance exam' Studio use case, when Milestones is enabled
    """
    [_inactivate_record(record) for record in internal.CourseContentMilestone.objects.filter(
        content_id=six.text_type(content_key),
        active=True
    )]


def delete_course_references(course_key):
    """
    Inactivates references to course keys within this app (ref: receivers.py and api.py)
    """
    [_inactivate_record(record) for record in internal.CourseMilestone.objects.filter(
        course_id=six.text_type(course_key),
        active=True
    )]

    [_inactivate_record(record) for record in internal.CourseContentMilestone.objects.filter(
        course_id=six.text_type(course_key),
        active=True
    )]



"""
Validators confirm the integrity of inbound information prior to a data.py handoff
"""
from __future__ import absolute_import, unicode_literals

import json
import six

from opaque_keys import InvalidKeyError
from opaque_keys.edx.keys import CourseKey, UsageKey

from .data import fetch_milestone_relationship_types


def course_key_is_valid(course_key):
    """
    Course key object validation
    """
    if course_key is None:
        return False
    try:
        CourseKey.from_string(six.text_type(course_key))
    except InvalidKeyError:
        return False
    return True


def content_key_is_valid(content_key):
    """
    Course module/content/usage key object validation
    """
    if content_key is None:
        return False
    try:
        UsageKey.from_string(six.text_type(content_key))
    except InvalidKeyError:
        return False
    return True


def course_content_milestone_requirements_is_valid(requirements):
    """
    CourseContentMilestone.requirements should be a valid JSON string

    Args:
        requirements: JSON serializable object containing requirements data

    Returns:
        bool: True if valid, otherwise False
    """
    try:
        json.dumps(requirements)
    except TypeError:
        return False
    return True


def milestone_data_is_valid(milestone_data):
    """
    Milestone data validation
    """
    if milestone_data is None:
        return False
    if 'id' in milestone_data and not milestone_data.get('id'):
        return False
    if 'name' in milestone_data and not milestone_data.get('name'):
        return False
    if 'namespace' in milestone_data and not milestone_data.get('namespace'):
        return False
    return True


def milestone_relationship_type_is_valid(name):
    """
    Milestone relationship type object validation
    """
    return name in list(fetch_milestone_relationship_types().values())


def user_is_valid(user):
    """
    User object validation
    """
    if user is None:
        return False
    if not user.get('id', 0):
        return False
    return True



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from __future__ import absolute_import
from django.db import models, migrations
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CourseContentMilestone',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('course_id', models.CharField(max_length=255, db_index=True)),
                ('content_id', models.CharField(max_length=255, db_index=True)),
                ('active', models.BooleanField(default=True)),
            ],
        ),
        migrations.CreateModel(
            name='CourseMilestone',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('course_id', models.CharField(max_length=255, db_index=True)),
                ('active', models.BooleanField(default=True)),
            ],
        ),
        migrations.CreateModel(
            name='Milestone',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('namespace', models.CharField(max_length=255, db_index=True)),
                ('name', models.CharField(max_length=255, db_index=True)),
                ('display_name', models.CharField(max_length=255)),
                ('description', models.TextField()),
                ('active', models.BooleanField(default=True)),
            ],
        ),
        migrations.CreateModel(
            name='MilestoneRelationshipType',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('name', models.CharField(unique=True, max_length=25, db_index=True)),
                ('description', models.TextField(blank=True)),
                ('active', models.BooleanField(default=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UserMilestone',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('user_id', models.IntegerField(db_index=True)),
                ('source', models.TextField(blank=True)),
                ('collected', models.DateTimeField(null=True, blank=True)),
                ('active', models.BooleanField(default=True)),
                ('milestone', models.ForeignKey(to='milestones.Milestone')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='milestone',
            unique_together=set([('namespace', 'name')]),
        ),
        migrations.AddField(
            model_name='coursemilestone',
            name='milestone',
            field=models.ForeignKey(to='milestones.Milestone'),
        ),
        migrations.AddField(
            model_name='coursemilestone',
            name='milestone_relationship_type',
            field=models.ForeignKey(to='milestones.MilestoneRelationshipType'),
        ),
        migrations.AddField(
            model_name='coursecontentmilestone',
            name='milestone',
            field=models.ForeignKey(to='milestones.Milestone'),
        ),
        migrations.AddField(
            model_name='coursecontentmilestone',
            name='milestone_relationship_type',
            field=models.ForeignKey(to='milestones.MilestoneRelationshipType'),
        ),
        migrations.AlterUniqueTogether(
            name='usermilestone',
            unique_together=set([('user_id', 'milestone')]),
        ),
        migrations.AlterUniqueTogether(
            name='coursemilestone',
            unique_together=set([('course_id', 'milestone')]),
        ),
        migrations.AlterUniqueTogether(
            name='coursecontentmilestone',
            unique_together=set([('course_id', 'content_id', 'milestone')]),
        ),
    ]



"""
Application-specific exception classes used throughout the implementation
"""
from __future__ import absolute_import, unicode_literals
from django.core.exceptions import ValidationError


class InvalidCourseKeyException(ValidationError):
    """ CourseKey validation exception class """


class InvalidContentKeyException(ValidationError):
    """ Course content/module/usage key validation exception class """


class InvalidCourseContentMilestoneRequirementsException(ValidationError):
    """ CourseContentMilestone.requirements validation exception class """


class InvalidMilestoneException(ValidationError):
    """ Milestone validation exception class """


class InvalidMilestoneRelationshipTypeException(ValidationError):
    """ Milestone Relationship Type validation exception class """


class InvalidUserException(ValidationError):
    """ User validation exception class """


def raise_exception(entity_type, entity, exception):
    """ Exception helper """
    raise exception(
        'The {} you have provided is not valid: {}'.format(entity_type, entity)
    )



"""
Admin module for milestones app
"""
from __future__ import absolute_import, unicode_literals
from django.contrib import admin

from milestones.models import (
    CourseContentMilestone,
    CourseMilestone,
    Milestone,
    MilestoneRelationshipType,
    UserMilestone,
)


admin.site.register(CourseContentMilestone)
admin.site.register(CourseMilestone)
admin.site.register(Milestone)
admin.site.register(MilestoneRelationshipType)
admin.site.register(UserMilestone)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from __future__ import absolute_import
from django.db import migrations, models

from milestones.data import fetch_milestone_relationship_types


def seed_relationship_types(apps, schema_editor):
    """Seed the relationship types."""
    MilestoneRelationshipType = apps.get_model("milestones", "MilestoneRelationshipType")
    for name in fetch_milestone_relationship_types().values():
        MilestoneRelationshipType.objects.get_or_create(
            name=name,
            description='Autogenerated milestone relationship type "{}"'.format(name),
        )


def delete_relationship_types(apps, schema_editor):
    """Clean up any relationships we made."""
    MilestoneRelationshipType = apps.get_model("milestones", "MilestoneRelationshipType")
    for name in fetch_milestone_relationship_types().values():
        MilestoneRelationshipType.objects.filter(name=name).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('milestones', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(seed_relationship_types, delete_relationship_types),
    ]



# -*- coding: utf-8 -*- pylint: disable=too-many-lines
# pylint: disable=invalid-name
# pylint: disable=too-many-public-methods
"""
Milestones API Module Test Cases
"""
from __future__ import absolute_import, unicode_literals

from opaque_keys.edx.keys import UsageKey

import milestones.api as api
import milestones.exceptions as exceptions
import milestones.tests.utils as utils
import six


class MilestonesApiTestCase(utils.MilestonesTestCaseMixin, utils.MilestonesTestCaseBase):
    """
    Main Test Case module for Milestones API
    """
    def setUp(self):
        """
        Milestones API Test Case scaffolding
        """
        super(MilestonesApiTestCase, self).setUp()
        self.test_milestone = api.add_milestone({
            'name': 'test_milestone',
            'display_name': 'Test Milestone',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Test Milestone Description',
        })
        self.relationship_types = api.get_milestone_relationship_types()

    def test_add_milestone(self):
        """ Unit Test: test_add_milestone"""
        with self.assertNumQueries(2):
            milestone = api.add_milestone({
                'name': 'local_milestone',
                'display_name': 'Local Milestone',
                'namespace': six.text_type(self.test_course_key),
                'description': 'Local Milestone Description'
            })
        self.assertGreater(milestone['id'], 0)

    def test_add_milestone_active_exists(self):
        """ Unit Test: test_add_milestone_active_exists"""
        milestone_data = {
            'name': 'local_milestone',
            'display_name': 'Local Milestone',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Local Milestone Description'
        }
        milestone = api.add_milestone(milestone_data)
        self.assertGreater(milestone['id'], 0)
        with self.assertNumQueries(1):
            milestone = api.add_milestone(milestone_data)

    def test_add_milestone_inactive_to_active(self):
        """ Unit Test: test_add_milestone_inactive_to_active """
        milestone_data = {
            'name': 'local_milestone',
            'display_name': 'Local Milestone',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Local Milestone Description'
        }
        milestone = api.add_milestone(milestone_data)
        self.assertGreater(milestone['id'], 0)
        api.remove_milestone(milestone['id'])

        with self.assertNumQueries(6):
            milestone = api.add_milestone(milestone_data)

    def test_add_milestone_inactive_milestone_with_relationships(self):
        """ Unit Test: test_add_milestone_inactive_milestone_with_relationships"""
        milestone_data = {
            'name': 'local_milestone',
            'display_name': 'Local Milestone',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Local Milestone Description'
        }
        milestone = api.add_milestone(milestone_data)
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            milestone
        )
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['FULFILLS'],
            milestone
        )
        api.add_user_milestone(self.serialized_test_user, milestone)
        self.assertGreater(milestone['id'], 0)
        api.remove_milestone(milestone['id'])

        with self.assertNumQueries(9):
            milestone = api.add_milestone(milestone_data)

    def test_add_milestone_inactive_milestone_with_relationships_propagate_false(self):
        """ Unit Test: test_add_milestone_inactive_milestone_with_relationships_propagate_false"""
        milestone_data = {
            'name': 'local_milestone',
            'display_name': 'Local Milestone',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Local Milestone Description'
        }
        milestone = api.add_milestone(milestone_data)
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            milestone
        )
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['FULFILLS'],
            milestone
        )
        api.add_user_milestone(self.serialized_test_user, milestone)
        self.assertGreater(milestone['id'], 0)
        api.remove_milestone(milestone['id'])

        with self.assertNumQueries(3):
            milestone = api.add_milestone(milestone_data, propagate=False)

    def test_add_milestone_invalid_data_throws_exceptions(self):
        """ Unit Test: test_add_milestone_invalid_namespaces_throw_exceptions"""
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidMilestoneException):
                api.add_milestone({
                    'name': 'local_milestone',
                    'display_name': 'Local Milestone',
                    'namespace': '',  # Should throw an exception
                    'description': 'Local Milestone Description'
                })

        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidMilestoneException):
                api.add_milestone({
                    'name': 'Local Milestone',  # Missing namespace should throw exception
                    'description': 'Local Milestone Description'
                })

        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidMilestoneException):
                api.add_milestone({
                    'name': '',  # Empty name should throw an exception on create
                    'namespace': 'fixed',
                    'description': 'Local Milestone Description 2'
                })

        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidMilestoneException):
                api.add_milestone({
                    'namespace': 'fixed',  # Missing name should throw exception
                    'description': 'Local Milestone Description 2'
                })

    def test_edit_milestone(self):
        """ Unit Test: test_edit_milestone"""
        self.test_milestone['name'] = 'Edited Milestone'

        with self.assertNumQueries(1):
            api.edit_milestone(self.test_milestone)

    def test_edit_milestone_invalid_data_throws_exceptions(self):
        """ Unit Test: test_edit_milestone_missing_namespace """
        self.test_milestone['namespace'] = ''
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidMilestoneException):
                api.edit_milestone(self.test_milestone)

        self.test_milestone['namespace'] = 'fixed'
        self.test_milestone['id'] = 0
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidMilestoneException):
                api.edit_milestone(self.test_milestone)

    def test_edit_milestone_bogus_milestone(self):
        """ Unit Test: test_edit_milestone_bogus_milestone """
        self.test_milestone['id'] = 12345
        self.test_milestone['namespace'] = 'bogus.milestones'
        with self.assertNumQueries(1):
            with self.assertRaises(exceptions.InvalidMilestoneException):
                api.edit_milestone(self.test_milestone)
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidMilestoneException):
                api.edit_milestone(None)

    def test_get_milestone(self):
        """ Unit Test: test_get_milestone"""
        with self.assertNumQueries(1):
            milestone = api.get_milestone(self.test_milestone['id'])
        self.assertEqual(milestone['name'], self.test_milestone['name'])
        self.assertEqual(milestone['namespace'], self.test_milestone['namespace'])
        self.assertEqual(milestone['description'], self.test_milestone['description'])

    def test_get_milestones(self):
        """ Unit Test: test_get_milestones """
        namespace = 'test_get_milestones'
        api.add_milestone({
            'display_name': 'Local Milestone 1',
            'name': 'local_milestone_1',
            'namespace': namespace,
            'description': 'Local Milestone 1 Description'
        })
        api.add_milestone({
            'display_name': 'Local Milestone 2',
            'name': 'local_milestone_2',
            'namespace': namespace,
            'description': 'Local Milestone 2 Description'
        })
        with self.assertNumQueries(1):
            milestones = api.get_milestones(namespace=namespace)
        self.assertEqual(len(milestones), 2)

    def test_get_milestone_invalid_milestone(self):
        """ Unit Test: test_get_milestone_invalid_milestone """
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidMilestoneException):
                api.get_milestone(None)
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidMilestoneException):
                api.get_milestone(0)

    def test_remove_milestone(self):
        """ Unit Test: test_remove_milestone """
        with self.assertNumQueries(5):
            api.remove_milestone(self.test_milestone['id'])
        with self.assertRaises(exceptions.InvalidMilestoneException):
            api.get_milestone(self.test_milestone['id'])

    def test_remove_milestone_bogus_milestone(self):
        """ Unit Test: test_remove_milestone_bogus_milestone """
        with self.assertNumQueries(5):
            api.remove_milestone(self.test_milestone['id'])

        with self.assertRaises(exceptions.InvalidMilestoneException):
            api.get_milestone(self.test_milestone['id'])

        # Do it again with the valid id to hit the exception workflow
        with self.assertNumQueries(4):
            api.remove_milestone(self.test_milestone['id'])

        with self.assertRaises(exceptions.InvalidMilestoneException):
            api.get_milestone(self.test_milestone['id'])

    def test_add_course_milestone(self):
        """ Unit Test: test_add_course_milestone """
        with self.assertNumQueries(3):
            api.add_course_milestone(
                self.test_course_key,
                self.relationship_types['REQUIRES'],
                self.test_milestone
            )
        requirer_milestones = api.get_course_milestones(
            self.test_course_key,
            self.relationship_types['REQUIRES']
        )
        self.assertEqual(len(requirer_milestones), 1)

        with self.assertNumQueries(3):
            api.add_course_milestone(
                self.test_prerequisite_course_key,
                self.relationship_types['FULFILLS'],
                self.test_milestone
            )
        fulfiller_milestones = api.get_course_milestones(
            self.test_prerequisite_course_key,
            self.relationship_types['FULFILLS'],
        )
        self.assertEqual(len(fulfiller_milestones), 1)

    def test_add_course_milestone_active_exists(self):
        """ Unit Test: test_add_course_milestone """
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        with self.assertNumQueries(2):
            api.add_course_milestone(
                self.test_course_key,
                self.relationship_types['REQUIRES'],
                self.test_milestone
            )

    def test_add_course_milestone_inactive_to_active(self):
        """ Unit Test: test_add_course_milestone """
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        api.remove_course_milestone(self.test_course_key, self.test_milestone)
        with self.assertNumQueries(3):
            api.add_course_milestone(
                self.test_course_key,
                self.relationship_types['REQUIRES'],
                self.test_milestone
            )

    def test_add_course_milestone_bogus_course_key(self):
        """ Unit Test: test_add_course_milestone_bogus_course_key """
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidCourseKeyException):
                api.add_course_milestone('12345667av', 'whatever', self.test_milestone)
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidCourseKeyException):
                api.add_course_milestone(None, 'whatever', self.test_milestone)

    def test_add_course_milestone_bogus_milestone_relationship_type(self):
        """ Unit Test: test_add_course_milestone_bogus_milestone_relationship_type """
        with self.assertNumQueries(1):
            with self.assertRaises(exceptions.InvalidMilestoneRelationshipTypeException):
                api.add_course_milestone(self.test_course_key, 'whatever', self.test_milestone)

    def test_get_course_milestones(self):
        """ Unit Test: test_get_course_milestones """
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        with self.assertNumQueries(2):
            requirer_milestones = api.get_course_milestones(
                self.test_course_key,
                self.relationship_types['REQUIRES']
            )
        self.assertEqual(len(requirer_milestones), 1)

    def test_get_course_milestones_with_invalid_relationship_type(self):
        """ Unit Test: test_get_course_milestones_with_invalid_relationship_type """
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        with self.assertNumQueries(1):
            requirer_milestones = api.get_course_milestones(
                self.test_course_key,
                'INVALID RELATIONSHIP TYPE'
            )
        self.assertEqual(len(requirer_milestones), 0)

    def test_get_course_unfulfilled_milestones(self):
        """ Unit Test: test_get_course_unfulfilled_milestones """
        namespace = 'test_get_milestones'
        milestone1 = api.add_milestone({
            'name': 'localmilestone1',
            'display_name': 'Local Milestone 1',
            'namespace': namespace,
            'description': 'Local Milestone 1 Description'
        })
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            milestone1
        )

        milestone2 = api.add_milestone({
            'name': 'localmilestone2',
            'display_name': 'Local Milestone 2',
            'namespace': namespace,
            'description': 'Local Milestone 2 Description'
        })
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            milestone2
        )

        # Confirm that the course has only two milestones, and that the User still needs to collect both
        course_milestones = api.get_course_milestones(self.test_course_key)
        self.assertEqual(len(course_milestones), 2)
        with self.assertNumQueries(2):
            required_milestones = api.get_course_required_milestones(
                self.test_course_key,
                self.serialized_test_user
            )

        # Link the User to Milestone 2 (this one is now 'collected')
        api.add_user_milestone(self.serialized_test_user, milestone2)
        user_milestones = api.get_user_milestones(self.serialized_test_user, namespace=namespace)
        self.assertEqual(len(user_milestones), 1)
        self.assertEqual(user_milestones[0]['id'], milestone2['id'])

        # Only Milestone 1 should be listed as 'required' for the course at this point
        with self.assertNumQueries(2):
            required_milestones = api.get_course_required_milestones(
                self.test_course_key,
                self.serialized_test_user
            )
        self.assertEqual(len(required_milestones), 1)
        self.assertEqual(required_milestones[0]['id'], milestone1['id'])

        # Link the User to Milestone 1 (this one is now 'collected', as well)
        api.add_user_milestone(self.serialized_test_user, milestone1)
        user_milestones = api.get_user_milestones(self.serialized_test_user, namespace=namespace)
        self.assertEqual(len(user_milestones), 2)

        # And there should be no more Milestones required for this User+Course
        with self.assertNumQueries(2):
            required_milestones = api.get_course_required_milestones(
                self.test_course_key,
                self.serialized_test_user
            )
        self.assertEqual(len(required_milestones), 0)

    def test_get_courses_milestones(self):
        """ Unit Test: test_get_courses_milestones """
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        api.add_course_milestone(
            self.test_prerequisite_course_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        local_milestone = api.add_milestone({
            'display_name': 'Local Milestone',
            'name': 'local_milestone',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Local Milestone Description'
        })
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['FULFILLS'],
            local_milestone
        )
        with self.assertNumQueries(2):
            requirer_milestones = api.get_courses_milestones(
                [self.test_course_key, self.test_prerequisite_course_key],
                self.relationship_types['REQUIRES']
            )
        self.assertEqual(len(requirer_milestones), 2)

        with self.assertNumQueries(1):
            requirer_milestones = api.get_courses_milestones(
                [self.test_course_key],
            )
        self.assertEqual(len(requirer_milestones), 2)

    def test_get_courses_milestones_with_invalid_relationship_type(self):
        """ Unit Test: test_get_courses_milestones_with_invalid_relationship_type """
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        api.add_course_milestone(
            self.test_prerequisite_course_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        local_milestone = api.add_milestone({
            'display_name': 'Local Milestone',
            'name': 'local_milestone',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Local Milestone Description'
        })
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['FULFILLS'],
            local_milestone
        )

        with self.assertNumQueries(1):
            requirer_milestones = api.get_courses_milestones(
                [self.test_course_key],
                'INVALID RELATIONSHIP TYPE'
            )
        self.assertEqual(len(requirer_milestones), 0)

    def test_remove_course_milestone(self):
        """ Unit Test: test_remove_course_milestone """
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        requirer_milestones = api.get_course_milestones(
            self.test_course_key,
            self.relationship_types['REQUIRES']
        )
        self.assertEqual(len(requirer_milestones), 1)
        with self.assertNumQueries(2):
            api.remove_course_milestone(self.test_course_key, self.test_milestone)
        requirer_milestones = api.get_course_milestones(self.test_course_key)
        self.assertEqual(len(requirer_milestones), 0)

    def test_remove_course_milestone_missing_milestone(self):
        """ Unit Test: test_remove_course_milestone_missing_milestone """
        with self.assertNumQueries(1):
            api.remove_course_milestone(self.test_course_key, self.test_milestone)
        requirer_milestones = api.get_course_milestones(self.test_course_key)
        self.assertEqual(len(requirer_milestones), 0)

    def test_add_course_content_milestone(self):
        """ Unit Test: test_add_course_content_milestone """
        with self.assertNumQueries(3):
            api.add_course_content_milestone(
                self.test_course_key,
                self.test_content_key,
                self.relationship_types['REQUIRES'],
                self.test_milestone
            )
        requirer_milestones = api.get_course_content_milestones(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES']
        )
        self.assertEqual(len(requirer_milestones), 1)
        with self.assertNumQueries(3):
            api.add_course_content_milestone(
                self.test_prerequisite_course_key,
                self.test_content_key,
                self.relationship_types['FULFILLS'],
                self.test_milestone,
                {'min_score': 80}
            )
        fulfiller_milestones = api.get_course_content_milestones(
            self.test_prerequisite_course_key,
            self.test_content_key,
            self.relationship_types['FULFILLS']
        )
        self.assertEqual(len(fulfiller_milestones), 1)

    def test_add_course_content_milestone_active_exists(self):
        """ Unit Test: test_add_course_content_milestone_active_exists """
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        with self.assertNumQueries(2):
            api.add_course_content_milestone(
                self.test_course_key,
                self.test_content_key,
                self.relationship_types['REQUIRES'],
                self.test_milestone
            )

    def test_add_course_content_milestone_inactive_to_active(self):
        """ Unit Test: test_add_course_content_milestone_inactive_to_active """
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        api.remove_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.test_milestone
        )
        with self.assertNumQueries(3):
            api.add_course_content_milestone(
                self.test_course_key,
                self.test_content_key,
                self.relationship_types['REQUIRES'],
                self.test_milestone
            )

    def test_add_course_content_milestone_invalid_requirements(self):
        """ Unit Test: test_add_course_content_milestone_invalid_requirements """
        with self.assertRaises(exceptions.InvalidCourseContentMilestoneRequirementsException):
            api.add_course_content_milestone(
                self.test_course_key,
                self.test_content_key,
                self.relationship_types['REQUIRES'],
                self.test_milestone,
                set()  # Not JSON serializable
            )

    def test_add_course_content_milestone_active_exists_update_requirements(self):
        """ Unit Test: test_add_course_content_milestone_active_exists_update_requirement """
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone,
            {'min_score': 60}
        )
        with self.assertNumQueries(3):
            api.add_course_content_milestone(
                self.test_course_key,
                self.test_content_key,
                self.relationship_types['REQUIRES'],
                self.test_milestone,
                {'min_score': 80}
            )
        milestone = api.get_course_content_milestones(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES']
        )
        for m in milestone:
            self.assertEqual(m['requirements'], {'min_score': 80})

    def test_add_course_content_milestone_inactive_to_active_update_requirements(self):
        """ Unit Test: test_add_course_content_milestone_inactive_to_active_update_requirement """
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone,
            {'min_score': 60}
        )
        api.remove_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.test_milestone
        )
        with self.assertNumQueries(3):
            api.add_course_content_milestone(
                self.test_course_key,
                self.test_content_key,
                self.relationship_types['REQUIRES'],
                self.test_milestone,
                {'min_score': 80}
            )
        milestone = api.get_course_content_milestones(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES']
        )
        for m in milestone:
            self.assertEqual(m['requirements'], {'min_score': 80})

    def test_add_course_content_milestone_bogus_content_key(self):
        """ Unit Test: test_add_course_content_milestone_bogus_content_key """
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidContentKeyException):
                api.add_course_content_milestone(
                    self.test_course_key,
                    '234290jweovsu',
                    'whatever',
                    self.test_milestone
                )
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidContentKeyException):
                api.add_course_content_milestone(
                    self.test_course_key,
                    None,
                    'whatever',
                    self.test_milestone
                )

    def test_add_course_content_milestone_bogus_milestone_relationship_type(self):
        """ Unit Test: test_add_course_content_milestone_bogus_milestone_relationship_type """
        with self.assertNumQueries(1):
            with self.assertRaises(exceptions.InvalidMilestoneRelationshipTypeException):
                api.add_course_content_milestone(
                    self.test_course_key,
                    self.test_content_key,
                    'whatever',
                    self.test_milestone
                )
        with self.assertNumQueries(1):
            with self.assertRaises(exceptions.InvalidMilestoneRelationshipTypeException):
                api.add_course_content_milestone(
                    self.test_course_key,
                    self.test_content_key,
                    None,
                    self.test_milestone
                )

    def test_get_course_content_milestones(self):
        """ Unit Test: test_get_course_content_milestones """
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        with self.assertNumQueries(2):
            requirer_milestones = api.get_course_content_milestones(
                self.test_course_key,
                self.test_content_key,
                self.relationship_types['REQUIRES']
            )
        self.assertEqual(len(requirer_milestones), 1)

    def test_get_course_content_milestones_with_unfulfilled_user_milestones(self):
        """ Unit Test: test_get_course_content_milestones_with_unfulfilled_user_milestones """
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        with self.assertNumQueries(2):
            requirer_milestones = api.get_course_content_milestones(
                self.test_course_key,
                self.test_content_key,
                self.relationship_types['REQUIRES'],
                {'id': self.test_user.id}
            )
        self.assertEqual(len(requirer_milestones), 1)

    def test_get_course_content_milestones_with_fulfilled_user_milestones(self):
        """ Unit Test: test_get_course_content_milestones_with_fulfilled_user_milestones """
        user = {'id': self.test_user.id}
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        api.add_user_milestone(user, self.test_milestone)
        with self.assertNumQueries(2):
            requirer_milestones = api.get_course_content_milestones(
                self.test_course_key,
                self.test_content_key,
                self.relationship_types['REQUIRES'],
                {'id': self.test_user.id}
            )
        self.assertEqual(len(requirer_milestones), 0)

    def test_get_course_content_milestones_without_course_key(self):
        """ Unit Test: test_get_course_content_milestones_without_course_key """
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        api.add_course_content_milestone(
            self.test_alternate_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        with self.assertNumQueries(2):
            requirer_milestones = api.get_course_content_milestones(
                None,
                self.test_content_key,
                self.relationship_types['REQUIRES']
            )
        self.assertEqual(len(requirer_milestones), 2)

    def test_get_course_content_milestones_without_content_key(self):
        """ Unit Test: test_get_course_content_milestones_without_content_key """
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_alternate_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        with self.assertNumQueries(2):
            requirer_milestones = api.get_course_content_milestones(
                self.test_course_key,
                None,
                self.relationship_types['REQUIRES']
            )
        self.assertEqual(len(requirer_milestones), 2)

    def test_get_course_content_milestones_with_invalid_relationship(self):
        """ Unit Test: test_get_course_content_milestones_with_invalid_relationship """
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_alternate_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        with self.assertNumQueries(1):
            requirer_milestones = api.get_course_content_milestones(
                self.test_course_key,
                self.test_content_key,
                'INVALID RELATIONSHIP TYPE'
            )
        self.assertEqual(len(requirer_milestones), 0)

    def test_remove_course_content_milestone(self):
        """ Unit Test: test_remove_course_content_milestone """
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        requirer_milestones = api.get_course_content_milestones(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES']
        )
        self.assertEqual(len(requirer_milestones), 1)
        with self.assertNumQueries(2):
            api.remove_course_content_milestone(
                self.test_course_key,
                self.test_content_key,
                self.test_milestone
            )
        requirer_milestones = api.get_course_content_milestones(
            self.test_course_key,
            self.test_content_key
        )
        self.assertEqual(len(requirer_milestones), 0)

    def test_remove_course_content_milestone_missing_milestone(self):
        """ Unit Test: test_remove_course_content_milestone_missing_milestone """
        with self.assertNumQueries(1):
            api.remove_course_content_milestone(
                self.test_course_key,
                self.test_content_key,
                self.test_milestone
            )
        requirer_milestones = api.get_course_content_milestones(
            self.test_course_key,
            self.test_content_key
        )
        self.assertEqual(len(requirer_milestones), 0)

    def test_add_user_milestone(self):
        """ Unit Test: test_add_user_milestone """
        with self.assertNumQueries(2):
            api.add_user_milestone(self.serialized_test_user, self.test_milestone)
        self.assertTrue(api.user_has_milestone(self.serialized_test_user, self.test_milestone))

    def test_add_user_milestone_active_exists(self):
        """ Unit Test: test_add_user_milestone """
        api.add_user_milestone(self.serialized_test_user, self.test_milestone)
        with self.assertNumQueries(1):
            api.add_user_milestone(self.serialized_test_user, self.test_milestone)
        self.assertTrue(api.user_has_milestone(self.serialized_test_user, self.test_milestone))

    def test_add_user_milestone_inactive_to_active(self):
        """ Unit Test: test_add_user_milestone """
        api.add_user_milestone(self.serialized_test_user, self.test_milestone)
        api.remove_user_milestone(self.serialized_test_user, self.test_milestone)
        with self.assertNumQueries(2):
            api.add_user_milestone(self.serialized_test_user, self.test_milestone)
        self.assertTrue(api.user_has_milestone(self.serialized_test_user, self.test_milestone))

    def test_add_user_milestone_bogus_user(self):
        """ Unit Test: test_add_user_milestone_bogus_user """
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidUserException):
                api.add_user_milestone({'identifier': 'abcd'}, self.test_milestone)
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidUserException):
                api.add_user_milestone(None, self.test_milestone)

    def test_get_user_milestones(self):
        """ Unit Test: test_get_user_milestones """
        with self.assertNumQueries(2):
            api.add_user_milestone(self.serialized_test_user, self.test_milestone)
        self.assertTrue(api.user_has_milestone(self.serialized_test_user, self.test_milestone))

    def test_remove_user_milestone(self):
        """ Unit Test: test_remove_user_milestone """
        api.add_user_milestone(self.serialized_test_user, self.test_milestone)
        self.assertTrue(api.user_has_milestone(self.serialized_test_user, self.test_milestone))
        with self.assertNumQueries(2):
            api.remove_user_milestone(self.serialized_test_user, self.test_milestone)
        self.assertFalse(api.user_has_milestone(self.serialized_test_user, self.test_milestone))

    def test_remove_user_milestone_missing_milestone(self):
        """ Unit Test: test_remove_user_milestone_missing_milestone """
        with self.assertNumQueries(1):
            api.remove_user_milestone(self.serialized_test_user, self.test_milestone)
        self.assertFalse(api.user_has_milestone(self.serialized_test_user, self.test_milestone))

    def test_user_has_milestone(self):
        """ Unit Test: test_user_has_milestone """
        api.add_user_milestone(self.serialized_test_user, self.test_milestone)
        with self.assertNumQueries(1):
            self.assertTrue(api.user_has_milestone(self.serialized_test_user, self.test_milestone))
        api.remove_user_milestone(self.serialized_test_user, self.test_milestone)
        with self.assertNumQueries(1):
            self.assertFalse(api.user_has_milestone(self.serialized_test_user, self.test_milestone))

    def test_remove_course_references(self):
        """ Unit Test: test_remove_course_references """
        # Add a course dependency on the test milestone
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        self.assertEqual(len(api.get_course_milestones(self.test_course_key)), 1)

        # Add a content fulfillment for the test milestone
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['FULFILLS'],
            self.test_milestone
        )
        self.assertEqual(
            len(api.get_course_content_milestones(self.test_course_key, self.test_content_key)), 1)

        # Remove the course dependency
        with self.assertNumQueries(4):
            api.remove_course_references(self.test_course_key)
        self.assertEqual(len(api.get_course_milestones(self.test_course_key)), 0)

    def test_remove_content_references(self):
        """ Unit Test: test_remove_content_references """
        # Add a course dependency on the test milestone
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            self.test_milestone
        )
        self.assertEqual(len(api.get_course_milestones(self.test_course_key)), 1)

        # Add a content fulfillment for the test milestone
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['FULFILLS'],
            self.test_milestone
        )
        milestones = api.get_course_content_milestones(self.test_course_key, self.test_content_key)
        self.assertEqual(len(milestones), 1)

        # Remove the content dependency
        with self.assertNumQueries(2):
            api.remove_content_references(self.test_content_key)
        milestones = api.get_course_content_milestones(self.test_course_key, self.test_content_key)
        self.assertEqual(len(milestones), 0)

    def test_milestones_fulfillment_paths_contains_special_characters(self):
        """
        Unit Test: test_get_course_milestones_fulfillment_paths works correctly when milestone have some special
        characters.
        """
        namespace = six.text_type(self.test_course_key)
        name = '�ťÉśt_Àübùr�'
        local_milestone_1 = api.add_milestone({
            'display_name': 'Local Milestone 1',
            'name': name,
            'namespace': namespace,
            'description': 'Local Milestone 1 Description'
        })

        # Specify the milestone requirements
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            local_milestone_1
        )

        # Specify the milestone fulfillments (via course and content)
        api.add_course_milestone(
            self.test_prerequisite_course_key,
            self.relationship_types['FULFILLS'],
            local_milestone_1
        )
        with self.assertNumQueries(4):
            paths = api.get_course_milestones_fulfillment_paths(
                self.test_course_key,
                self.serialized_test_user
            )

        # Set up the key values we'll use to access/assert the response
        milestone_key_1 = '{}.{}'.format(local_milestone_1['namespace'], local_milestone_1['name'])
        self.assertEqual(len(paths[milestone_key_1]['courses']), 1)

    def test_get_course_milestones_fulfillment_paths(self):  # pylint: disable=too-many-statements
        """
        Unit Test: test_get_course_milestones_fulfillment_paths
        """
        # Create three milestones in order tto cover all logical branches
        namespace = six.text_type(self.test_course_key)
        local_milestone_1 = api.add_milestone({
            'display_name': 'Local Milestone 1',
            'name': 'local_milestone_1',
            'namespace': namespace,
            'description': 'Local Milestone 1 Description'
        })
        local_milestone_2 = api.add_milestone({
            'display_name': 'Local Milestone 2',
            'name': 'local_milestone_2',
            'namespace': namespace,
            'description': 'Local Milestone 2 Description'
        })
        local_milestone_3 = api.add_milestone({
            'display_name': 'Local Milestone 3',
            'name': 'local_milestone_3',
            'namespace': namespace,
            'description': 'Local Milestone 3 Description'
        })

        # Specify the milestone requirements
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            local_milestone_1
        )
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            local_milestone_2
        )
        api.add_course_milestone(
            self.test_course_key,
            self.relationship_types['REQUIRES'],
            local_milestone_3
        )

        # Specify the milestone fulfillments (via course and content)
        api.add_course_milestone(
            self.test_prerequisite_course_key,
            self.relationship_types['FULFILLS'],
            local_milestone_1
        )
        api.add_course_milestone(
            self.test_prerequisite_course_key,
            self.relationship_types['FULFILLS'],
            local_milestone_2
        )
        api.add_course_content_milestone(
            self.test_course_key,
            UsageKey.from_string('i4x://the/content/key/123456789'),
            self.relationship_types['FULFILLS'],
            local_milestone_2
        )
        api.add_course_content_milestone(
            self.test_course_key,
            UsageKey.from_string('i4x://the/content/key/123456789'),
            self.relationship_types['FULFILLS'],
            local_milestone_3
        )
        api.add_course_content_milestone(
            self.test_course_key,
            UsageKey.from_string('i4x://the/content/key/987654321'),
            self.relationship_types['FULFILLS'],
            local_milestone_3
        )

        # Confirm the starting state for this test (user has no milestones, course requires three)
        self.assertEqual(
            len(api.get_user_milestones(self.serialized_test_user, namespace=namespace)), 0)
        self.assertEqual(
            len(api.get_course_required_milestones(self.test_course_key, self.serialized_test_user)),
            3
        )
        # Check the possible fulfillment paths for the milestones for this course
        with self.assertNumQueries(8):
            paths = api.get_course_milestones_fulfillment_paths(
                self.test_course_key,
                self.serialized_test_user
            )

        # Set up the key values we'll use to access/assert the response
        milestone_key_1 = '{}.{}'.format(local_milestone_1['namespace'], local_milestone_1['name'])
        milestone_key_2 = '{}.{}'.format(local_milestone_2['namespace'], local_milestone_2['name'])
        milestone_key_3 = '{}.{}'.format(local_milestone_3['namespace'], local_milestone_3['name'])

        # First round of assertions
        self.assertEqual(len(paths[milestone_key_1]['courses']), 1)
        self.assertIsNone(paths[milestone_key_1].get('content'))
        self.assertEqual(len(paths[milestone_key_2]['courses']), 1)
        self.assertEqual(len(paths[milestone_key_2]['content']), 1)
        self.assertIsNone(paths[milestone_key_3].get('courses'))
        self.assertEqual(len(paths[milestone_key_3]['content']), 2)

        # Collect the first milestone (two should remain)
        api.add_user_milestone(self.serialized_test_user, local_milestone_1)
        self.assertEqual(
            len(api.get_user_milestones(self.serialized_test_user, namespace=namespace)), 1)
        self.assertEqual(
            len(api.get_course_required_milestones(self.test_course_key, self.serialized_test_user)),
            2
        )
        # Check the remaining fulfillment paths for the milestones for this course
        with self.assertNumQueries(6):
            paths = api.get_course_milestones_fulfillment_paths(
                self.test_course_key,
                self.serialized_test_user
            )
        self.assertIsNone(paths.get(milestone_key_1))
        self.assertEqual(len(paths[milestone_key_2]['courses']), 1)
        self.assertEqual(len(paths[milestone_key_2]['content']), 1)
        self.assertIsNone(paths[milestone_key_3].get('courses'))
        self.assertEqual(len(paths[milestone_key_3]['content']), 2)

        # Collect the second milestone (one should remain)
        api.add_user_milestone(self.serialized_test_user, local_milestone_2)
        self.assertEqual(
            len(api.get_user_milestones(self.serialized_test_user, namespace=namespace)), 2)
        self.assertEqual(
            len(api.get_course_required_milestones(self.test_course_key, self.serialized_test_user)),
            1
        )
        # Check the remaining fulfillment paths for the milestones for this course
        with self.assertNumQueries(4):
            paths = api.get_course_milestones_fulfillment_paths(
                self.test_course_key,
                self.serialized_test_user
            )
        self.assertIsNone(paths.get(milestone_key_1))
        self.assertIsNone(paths.get(milestone_key_2))
        self.assertIsNone(paths[milestone_key_3].get('courses'))
        self.assertEqual(len(paths[milestone_key_3]['content']), 2)

        # Collect the third milestone
        api.add_user_milestone(self.serialized_test_user, local_milestone_3)
        self.assertEqual(
            len(api.get_user_milestones(self.serialized_test_user, namespace=namespace)), 3)
        self.assertEqual(
            len(api.get_course_required_milestones(self.test_course_key, self.serialized_test_user)),
            0
        )
        # Check the remaining fulfillment paths for the milestones for this course
        with self.assertNumQueries(2):
            paths = api.get_course_milestones_fulfillment_paths(
                self.test_course_key,
                self.serialized_test_user
            )
        self.assertIsNone(paths.get(milestone_key_1))
        self.assertIsNone(paths.get(milestone_key_2))
        self.assertIsNone(paths.get(milestone_key_3))



"""
Views are listeners connected to the Django HTTP request
pipeline.  When they observe a corresponding URL request they run the
code contained within their definition.

In this particular application, the views simply hand-off to the
orchestration layer, which manages the application's workflows.
"""

from __future__ import unicode_literals



# pylint: disable=no-member,too-many-public-methods
"""
Utility module for Milestones test cases
"""
from __future__ import absolute_import, unicode_literals
from django.contrib.auth.models import User
from django.test import TestCase
from opaque_keys.edx.keys import CourseKey, UsageKey
from milestones.models import MilestoneRelationshipType


class MilestonesTestCaseBase(TestCase):
    """
    Parent/Base class for Milestones test cases
    """

    def setUp(self):
        """
        Helper method for test case scaffolding
        """
        super(MilestonesTestCaseBase, self).setUp()
        self.test_course_key = CourseKey.from_string('the/course/key')
        self.test_alternate_course_key = CourseKey.from_string('the/alternate_course/key')
        self.test_prerequisite_course_key = CourseKey.from_string('the/prerequisite/key')
        self.test_content_key = UsageKey.from_string('i4x://the/content/key/12345678')
        self.test_alternate_content_key = UsageKey.from_string('i4x://the/alternate_content/key/12345678')
        self.test_user = User.objects.create_user(
            first_name='Test',
            last_name='User',
            email='test_user@edx.org',
            username='test_user',
            password='ABcd12!@'
        )
        self.serialized_test_user = self.test_user.__dict__


class MilestonesTestCaseMixin(TestCase):
    """
    TestCase mixin for loading initial milestones data
    """
    def setUp(self):
        super(MilestonesTestCaseMixin, self).setUp()
        MilestoneRelationshipType.objects.get_or_create(name='requires')
        MilestoneRelationshipType.objects.get_or_create(name='fulfills')



"""
Test for the xBlock service
"""

from __future__ import absolute_import, unicode_literals
import unittest
import types

from milestones.services import MilestonesService
from milestones import api as milestones_api


class TestMilestonesService(unittest.TestCase):  # pylint: disable=too-many-public-methods
    """
    Tests for MilestonesService
    """
    def test_basic(self):
        """
        See if the MilestonesService exposes the expected methods
        """

        service = MilestonesService()

        for attr_name in dir(milestones_api):
            attr = getattr(milestones_api, attr_name, None)
            if isinstance(attr, types.FunctionType) and not attr_name.startswith('_'):
                if attr_name in MilestonesService.REQUESTED_FUNCTIONS:
                    self.assertTrue(hasattr(service, attr_name))
                else:
                    self.assertFalse(hasattr(service, attr_name))

    def test_singleton(self):
        """
        Test to make sure the MilestonesService is a singleton.
        """
        service1 = MilestonesService()
        service2 = MilestonesService()
        self.assertIs(service1, service2)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from __future__ import absolute_import
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('milestones', '0003_coursecontentmilestone_requirements'),
    ]

    operations = [
        migrations.AlterField(
            model_name='coursecontentmilestone',
            name='active',
            field=models.BooleanField(default=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='coursemilestone',
            name='active',
            field=models.BooleanField(default=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='milestone',
            name='active',
            field=models.BooleanField(default=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='usermilestone',
            name='active',
            field=models.BooleanField(default=True, db_index=True),
        ),
    ]



"""
api.py is an interface module for Python-level integration with the
Milestones app.

In the current incarnation of this particular application, the API
operations are equivalent to the orchestration layer, which manages
the application's workflows.

Helplful Hint:  When modeling Milestones, I've found that it's helpful
to first consider the process for fulfilling the Milestone, and then
consider the process for requiring the Milestone.

Note the terminology difference at this layer vs. Data -- add/edit/get/remove
"""
from __future__ import absolute_import, unicode_literals
from . import data
from . import exceptions
from . import validators


# PRIVATE/INTERNAL FUNCTIONS

def _validate_course_key(course_key):
    """ Validation helper """
    if not validators.course_key_is_valid(course_key):
        exceptions.raise_exception(
            "CourseKey",
            course_key,
            exceptions.InvalidCourseKeyException
        )


def _validate_content_key(content_key):
    """ Validation helper """
    if not validators.content_key_is_valid(content_key):
        exceptions.raise_exception(
            "ContentKey",
            content_key,
            exceptions.InvalidContentKeyException
        )


def _validate_course_content_milestone_requirements(requirements):
    """ CourseContentMilestone.requirements field validation helper """
    if not validators.course_content_milestone_requirements_is_valid(requirements):
        exceptions.raise_exception(
            "CourseContentMilestone.requirements",
            requirements,
            exceptions.InvalidCourseContentMilestoneRequirementsException
        )


def _validate_milestone_data(milestone):
    """ Validation helper """
    if not validators.milestone_data_is_valid(milestone):
        exceptions.raise_exception(
            "Milestone",
            milestone,
            exceptions.InvalidMilestoneException
        )


def _validate_user(user):
    """ Validation helper """
    if not validators.user_is_valid(user):
        exceptions.raise_exception(
            "User",
            user,
            exceptions.InvalidUserException
        )


# PUBLIC FUNCTIONS
def get_milestone_relationship_types():
    """
    Exposes the available relationship type choices without exposing the data layer
    """
    return data.fetch_milestone_relationship_types()


def add_milestone(milestone, propagate=True):
    """
    Passes a new milestone to the data layer for storage

    Arguments:
        milestone (dict): The milestone to persist
        propagate (bool): False to prevent reactivation of soft-deleted milestone relationships

    Returns:
        dict: The persisted milestone dict
    """
    _validate_milestone_data(milestone)
    milestone = data.create_milestone(milestone, propagate)
    return milestone


def edit_milestone(milestone):
    """
    Passes an updated milestone to the data layer for storage
    """
    _validate_milestone_data(milestone)
    return data.update_milestone(milestone)


def get_milestone(milestone_id):
    """
    Retrieves the specified milestone
    """
    return data.fetch_milestone(milestone_id)


def get_milestones(namespace):
    """
    Retrieves a set of milestones by namespace
    (no other way to really do it right now)
    """
    milestone = {
        'namespace': namespace,
    }
    return data.fetch_milestones(milestone)


def remove_milestone(milestone_id):
    """
    Removes the specified milestone
    """
    milestone = {
        'id': milestone_id,
    }
    data.delete_milestone(milestone)


def add_course_milestone(course_key, relationship, milestone):
    """
    Adds a course-milestone link to the system
    'relationship': string value (eg: 'requires')
    """
    _validate_course_key(course_key)
    _validate_milestone_data(milestone)
    data.create_course_milestone(
        course_key=course_key,
        relationship=relationship,
        milestone=milestone
    )


def get_course_milestones(course_key, relationship=None):
    """
    Retrieves the set of milestones for a given course
    'relationship': optional filter on milestone relationship type (string, eg: 'requires')
    Returns an array of dicts containing milestones
    """
    _validate_course_key(course_key)

    try:
        milestones = data.fetch_courses_milestones(course_keys=[course_key], relationship=relationship)
    except exceptions.InvalidMilestoneRelationshipTypeException:
        milestones = []

    return milestones


def get_course_required_milestones(course_key, user):
    """
    Retrieves the set of required milestones for a given course that a user has not yet collected
    """
    _validate_course_key(course_key)
    _validate_user(user)
    required_milestones = data.fetch_courses_milestones(
        [course_key],
        get_milestone_relationship_types()['REQUIRES'],
        user
    )
    return required_milestones


def get_course_milestones_fulfillment_paths(course_key, user):
    """
    Returns a collection composed of the possible fulfillment/collection options/opportunites
    Inputs:
        course_key: string representation of a CourseKey -- eg, unicode(course.id)
        user: dictionary representation of a User object ('id' field is required)
    Output:
        A complex dict-of-dicts containing the currently-required milestones for the
        specified course+user combo, and the options for collecting them
        (courses and/or content to complete which leads to their fulfillment)
        {
            'milestone_1': {
                'content': [
                    u'i4x://the/content/key/123456789',
                    u'i4x://the/content/key/987654321'
                ]
            },
            'milestone_2': {
                'courses': [
                    u'the/prerequisite/course_key'
                ]
            },
            'milestone_3': {
                'content': [
                    u'i4x://the/content/key/123456789'
                ],
                'courses': [
                    u'the/prerequisite/course_key'
                ]
            }
        }
    """
    _validate_course_key(course_key)
    _validate_user(user)

    # Retrieve the outstanding milestones for this course, for this user
    required_milestones = data.fetch_courses_milestones(
        [course_key],
        get_milestone_relationship_types()['REQUIRES'],
        user
    )

    # Build the set of fulfillment paths for the outstanding milestones
    fulfillment_paths = {}
    for milestone in required_milestones:
        dict_key = '{}.{}'.format(milestone['namespace'], milestone['name'])
        fulfillment_paths[dict_key] = {}
        milestone_courses = data.fetch_milestone_courses(
            milestone,
            get_milestone_relationship_types()['FULFILLS']
        )
        if milestone_courses:
            fulfillment_paths[dict_key]['courses'] = [
                milestone['course_id'] for milestone in milestone_courses]
        milestone_course_content = data.fetch_milestone_course_content(
            milestone,
            get_milestone_relationship_types()['FULFILLS']
        )
        if milestone_course_content:
            fulfillment_paths[dict_key]['content'] = [
                milestone['content_id'] for milestone in milestone_course_content]
    return fulfillment_paths


def get_courses_milestones(course_keys, relationship=None, user=None):
    """
    Retrieves the set of milestones for list of courses
    'relationship': optional filter on milestone relationship type (string, eg: 'requires')
    'user': optional filter to constrain the set to those milestones which a user has already collected
    Returns an array of dicts containing milestones
    """
    [_validate_course_key(course_key) for course_key in course_keys]  # pylint: disable=expression-not-assigned
    try:
        milestones = data.fetch_courses_milestones(
            course_keys=course_keys,
            relationship=relationship,
            user=user
        )
    except exceptions.InvalidMilestoneRelationshipTypeException:
        milestones = []

    return milestones


def remove_course_milestone(course_key, milestone):
    """
    Removes the specfied milestone from the specified course
    """
    _validate_course_key(course_key)
    _validate_milestone_data(milestone)
    return data.delete_course_milestone(course_key=course_key, milestone=milestone)


def add_course_content_milestone(course_key, content_key, relationship, milestone, requirements=None):
    """
    Adds a course-content-milestone link to the system

    Arguments:
        course_key (CourseKey|str): CourseKey of the course containing the content
        content_key (UsageKey|str): UsageKey of the content
        relationship (str): The type of relationship that the content shares with the milestone (e.g. 'requires')
        milestone (dict): The milestone which the given content relates to
        requirements (dict): Data that is used to determine user fulfillment of the milestone

    Returns:
        None
    """
    _validate_course_key(course_key)
    _validate_content_key(content_key)
    _validate_milestone_data(milestone)
    _validate_course_content_milestone_requirements(requirements)
    data.create_course_content_milestone(
        course_key=course_key,
        content_key=content_key,
        relationship=relationship,
        milestone=milestone,
        requirements=requirements
    )


def get_course_content_milestones(course_key=None, content_key=None, relationship=None, user=None):
    """
    Retrieves the set of milestones related to course content

    Arguments:
        course_key (CourseKey|str): CourseKey of the course containing the content
        content_key (UsageKey|str): UsageKey of the content
        relationship (str): The type of relationship that the content shares with the milestone (e.g. 'requires')
        user (dict): Dict containing at least an 'id' key mapped to a user id

    Returns:
        list: List of milestone dicts
    """
    if course_key is not None:
        _validate_course_key(course_key)
    if content_key is not None:
        _validate_content_key(content_key)
    try:
        milestones = data.fetch_course_content_milestones(
            course_key=course_key,
            content_key=content_key,
            relationship=relationship,
            user=user
        )
    except exceptions.InvalidMilestoneRelationshipTypeException:
        milestones = []

    return milestones


def remove_course_content_milestone(course_key, content_key, milestone):
    """
    Removes the specified milestone from the specified course content module
    """
    _validate_course_key(course_key)
    _validate_content_key(content_key)
    _validate_milestone_data(milestone)
    return data.delete_course_content_milestone(
        course_key=course_key,
        content_key=content_key,
        milestone=milestone
    )


def add_user_milestone(user, milestone):
    """
    Adds a new User-Milestone relationship to the system
    """
    _validate_user(user)
    _validate_milestone_data(milestone)
    data.create_user_milestone(user, milestone)


def get_user_milestones(user, namespace):
    """
    Retrieves the set of milestones for a given user
    Returns an array of dicts -- must provide the 'namespace' filter (or else!!!)
    """
    _validate_user(user)
    milestone = {'namespace': namespace}
    return data.fetch_user_milestones(user, milestone)


def user_has_milestone(user, milestone):
    """
    A helper/convenience method to check for a specific user-milestone link
    """
    _validate_user(user)
    _validate_milestone_data(milestone)
    return True if data.fetch_user_milestones(user, milestone) else False


def remove_user_milestone(user, milestone):
    """
    Removes the specified User-Milestone link from the system
    """
    _validate_user(user)
    _validate_milestone_data(milestone)
    return data.delete_user_milestone(user, milestone)


def remove_course_references(course_key):
    """
    Removes course references from application state
    See edx-platform/lms/djangoapps/courseware/management/commands/delete_course_references.py
    """
    _validate_course_key(course_key)
    data.delete_course_references(course_key)


def remove_content_references(content_key):
    """
    Removes content references from application state
    See edx-platform/cms/djangoapps/contentstore/views/entrance_exam.py:_delete_entrance_exam
    """
    _validate_content_key(content_key)
    data.delete_content_references(content_key)



"""
Data layer serialization operations.  Converts querysets to simple
python containers (mainly arrays and dicts).
"""
from __future__ import absolute_import, unicode_literals
import json

from . import models


def serialize_milestone(milestone):
    """
    Milestone object-to-dict serialization
    """
    return {
        'id': milestone.id,
        'name': milestone.name,
        'display_name': milestone.display_name,
        'namespace': milestone.namespace,
        'description': milestone.description
    }


def serialize_milestone_with_course(course_milestone):
    """
    CourseMilestone serialization (composite object)
    """
    return {
        'id': course_milestone.milestone.id,
        'name': course_milestone.milestone.name,
        'display_name': course_milestone.milestone.display_name,
        'namespace': course_milestone.milestone.namespace,
        'description': course_milestone.milestone.description,
        'course_id': course_milestone.course_id
    }


def serialize_milestone_with_course_content(course_content_milestone):
    """
    CourseContentMilestone serialization (composite object)
    """
    return {
        'id': course_content_milestone.milestone.id,
        'name': course_content_milestone.milestone.name,
        'display_name': course_content_milestone.milestone.display_name,
        'namespace': course_content_milestone.milestone.namespace,
        'description': course_content_milestone.milestone.description,
        'course_id': course_content_milestone.course_id,
        'content_id': course_content_milestone.content_id,
        'requirements': deserialize_requirements(course_content_milestone.requirements)
    }


def serialize_milestones(milestones):
    """
    Milestone serialization
    Converts list of objects to list of dicts
    """
    return [serialize_milestone(milestone) for milestone in milestones]


def deserialize_milestone(milestone_dict):
    """
    Milestone dict-to-object serialization
    """
    return models.Milestone(
        id=milestone_dict.get('id'),
        name=milestone_dict.get('name', ''),
        display_name=milestone_dict.get('display_name', ''),
        namespace=milestone_dict.get('namespace', ''),
        description=milestone_dict.get('description', '')
    )


def serialize_requirements(requirements):
    """
    Convert JSON serializable object to string
    """
    if requirements is not None:
        requirements = json.dumps(requirements)
    return requirements


def deserialize_requirements(requirements):
    """
    Convert JSON string to object
    """
    if requirements is not None:
        requirements = json.loads(requirements)
    else:
        requirements = {}
    return requirements



#!/usr/bin/env python

from __future__ import absolute_import, unicode_literals
from setuptools import setup, find_packages
from milestones import __version__ as VERSION

setup(
    name='edx-milestones',
    version=VERSION,
    description='Significant events module for Open edX',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    author='edX',
    url='https://github.com/edx/edx-milestones',
    license='AGPL',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
    ],
    packages=find_packages(exclude=["tests"]),
    install_requires=[
        "django>=1.8,<2.0",
        "django-model-utils",
        "edx-opaque-keys>=0.2.1,<1.0.0",
        "six",
    ],
    tests_require=[
        "coverage==4.5.3",
        "nose==1.3.3",
        "httpretty==0.8.0",
        "pep8==1.5.7",
        "pylint==1.2.1",
        "pep257==0.3.2"
    ]
)



# pylint: disable=invalid-name
# pylint: disable=too-many-public-methods
"""
Milestones Data Module Test Cases

Note: 'Unit Test: ' labels are output to the console during test runs
"""
from __future__ import absolute_import, unicode_literals
import milestones.api as api
import milestones.data as data
import milestones.exceptions as exceptions
import milestones.tests.utils as utils
import six


class MilestonesDataTestCase(utils.MilestonesTestCaseBase):
    """
    Main Test Case module for Milestones Data
    Many of the module operations are covered indirectly via the test_api.py test suite
    So at the moment we're mainly focused on hitting the corner cases with this suite
    """
    def setUp(self):
        """
        Milestones Data Test Case scaffolding
        """
        super(MilestonesDataTestCase, self).setUp()
        self.relationship_types = api.get_milestone_relationship_types()

    def test_fetch_courses_milestones_invalid_milestone_relationship_type(self):
        """ Unit Test: test_fetch_courses_milestones_invalid_milestone_relationship_type"""
        milestone1 = api.add_milestone({
            'display_name': 'Test Milestone',
            'name': 'test_milestone',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Test Milestone Description',
        })
        api.add_course_milestone(self.test_course_key, self.relationship_types['REQUIRES'], milestone1)
        milestone2 = api.add_milestone({
            'display_name': 'Test Milestone 2',
            'name': 'test_milestone_2',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Test Milestone Description 2',
        })
        api.add_course_milestone(self.test_course_key, self.relationship_types['FULFILLS'], milestone2)
        with self.assertRaises(exceptions.InvalidMilestoneRelationshipTypeException):
            data.fetch_courses_milestones(
                [self.test_course_key, ],
                'invalid_relationshipppp',
                milestone1
            )

    def test_fetch_course_content_milestones_invalid_milestone_relationship_type(self):
        """ Unit Test: test_fetch_course_content_milestones_invalid_milestone_relationship_type"""
        milestone1 = api.add_milestone({
            'display_name': 'Test Milestone',
            'name': 'test_milestone',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Test Milestone Description',
        })
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            milestone1
        )
        milestone2 = api.add_milestone({
            'display_name': 'Test Milestone 2',
            'name': 'test_milestone2',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Test Milestone Description 2',
        })
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['FULFILLS'],
            milestone2
        )
        with self.assertRaises(exceptions.InvalidMilestoneRelationshipTypeException):
            data.fetch_course_content_milestones(
                self.test_course_key,
                self.test_content_key,
                'invalid_relationshipppp'
            )

    def test_fetch_courses_milestones_invalid_milestone(self):
        """ Unit Test: test_fetch_courses_milestones_invalid_milestone"""
        with self.assertRaises(exceptions.InvalidMilestoneException):
            data.fetch_milestones(milestone=None)
        with self.assertRaises(exceptions.InvalidMilestoneException):
            data.fetch_milestones(milestone={})

    def test_fetch_milestones_invalid_milestone_namespace(self):
        """ Unit Test: test_fetch_milestones_invalid_milestone_namespace"""
        milestones = data.fetch_milestones(milestone={'namespace': "some.namespace"})
        self.assertEqual(len(milestones), 0)

    def test_fetch_course_content_milestones_null_keys(self):
        """ Unit Test: test_fetch_course_content_milestones_null_keys"""
        namespace = '{}.entrance_exams'.format(six.text_type(self.test_course_key))
        milestone1 = api.add_milestone({
            'display_name': 'Test Milestone',
            'name': 'test_milestone',
            'namespace': namespace,
            'description': 'Test Milestone Description',
        })
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['REQUIRES'],
            milestone1
        )
        milestones = data.fetch_milestones(milestone={'namespace': namespace})
        self.assertEqual(len(milestones), 1)

        ccms = data.fetch_course_content_milestones(
            content_key=self.test_content_key,
            course_key=None,
            relationship=None
        )
        self.assertEqual(len(ccms), 1)

        ccms = data.fetch_course_content_milestones(
            content_key=None,
            course_key=self.test_course_key,
            relationship=None
        )
        self.assertEqual(len(ccms), 1)

    def test_fetch_milestone_courses_no_relationship_type(self):
        """ Unit Test: test_fetch_milestone_courses_no_relationship_type"""
        milestone1 = api.add_milestone({
            'display_name': 'Test Milestone',
            'name': 'test_milestone',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Test Milestone Description',
        })
        api.add_course_milestone(self.test_course_key, 'fulfills', milestone1)
        self.assertEqual(len(data.fetch_milestone_courses(milestone1)), 1)

    def test_fetch_milestone_course_content_no_relationship_type(self):
        """ Unit Test: test_fetch_milestone_course_content_no_relationship_type"""
        milestone1 = api.add_milestone({
            'display_name': 'Test Milestone',
            'name': 'test_milestone',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Test Milestone Description',
        })
        api.add_course_content_milestone(
            self.test_course_key,
            self.test_content_key,
            self.relationship_types['FULFILLS'],
            milestone1
        )
        self.assertEqual(len(data.fetch_milestone_course_content(milestone1)), 1)

    def test_fetch_user_milestones_missing_match_criteria_throws_exception(self):
        """ Unit Test: test_fetch_user_milestones_missing_match_criteria_throws_exception """
        milestone1 = api.add_milestone({
            'display_name': 'Test Milestone',
            'name': 'test_milestone',
            'namespace': six.text_type(self.test_course_key),
            'description': 'Test Milestone Description',
        })
        api.add_user_milestone(self.serialized_test_user, milestone1)

        with self.assertRaises(exceptions.InvalidMilestoneException):
            data.fetch_user_milestones(self.serialized_test_user, {})

        # Ensure we cover all remaining logical branches per coverage.py
        data.fetch_user_milestones(
            self.serialized_test_user,
            {'id': milestone1['id']}
        )
